from datetime import date, datetime, time, timedelta, timezone
from typing import Any
import json
import sys

from dateutil.parser import parse
import numpy as np
import pandas as pd


class SarusJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, np.ndarray):
            return {"_type": "numpy.ndarray", "data": obj.tolist()}
        elif isinstance(obj, pd.DataFrame):
            return {
                "_type": "pandas.DataFrame",
                "data": obj.to_json(date_format="iso"),
            }
        elif isinstance(obj, pd.Series):
            return {
                "_type": "pandas.Series",
                "data": obj.to_json(date_format="iso"),
            }
        elif isinstance(obj, pd.Timestamp):
            return {"_type": "pandas.Timestamp", "data": obj.isoformat()}
        elif isinstance(obj, datetime):
            return {"_type": "datetime", "data": obj.isoformat()}
        elif isinstance(obj, timedelta):
            return {"_type": "timedelta", "data": obj.total_seconds()}
        elif isinstance(obj, timezone):
            utcoffset_result = obj.utcoffset(None)
            if utcoffset_result is not None:
                return {
                    "_type": "timezone",
                    "data": utcoffset_result.total_seconds(),
                }
            else:
                raise ValueError("Invalid timezone object")
        elif isinstance(obj, time):
            return {"_type": "time", "data": obj.isoformat()}
        elif isinstance(obj, date):
            return {"_type": "date", "data": obj.isoformat()}
        elif isinstance(obj, np.generic):
            if np.issubdtype(obj, np.complexfloating):
                complex_obj = obj.astype(np.complex128)
                return {
                    "_type": "numpy.complex",
                    "data": {
                        '"real"': complex_obj.real,
                        '"imag"': complex_obj.imag,
                    },
                }
            else:
                return {"_type": "numpy.generic", "data": obj.item()}
        elif isinstance(obj, pd.MultiIndex):
            return {
                "_type": "pandas.MultiIndex",
                "data": obj.tolist(),
                "names": obj.names,
                "levels": [level.tolist() for level in obj.levels],
                "codes": list(obj.codes),
            }
        elif isinstance(obj, pd.Index):
            return {
                "_type": "pandas.Index",
                "class": type(obj).__name__,
                "data": obj.tolist(),
                "dtype": str(obj.dtype),
            }
        elif isinstance(obj, pd.Period):
            return {
                "_type": "pandas.Period",
                "data": str(obj),
                "freq": obj.freqstr,
            }
        elif isinstance(obj, pd.Timedelta):
            return {"_type": "pandas.Timedelta", "data": obj.value}
        elif isinstance(obj, pd.Interval):
            return {"_type": "pandas.Interval", "data": (obj.left, obj.right)}
        elif isinstance(obj, pd.Categorical):
            return {
                "_type": "pandas.Categorical",
                "data": obj.tolist(),
                "categories": obj.categories.tolist(),
                "ordered": obj.ordered,
            }
        elif isinstance(obj, type):
            return {
                "_type": "class",
                "data": {'"name"': obj.__name__, '"module"': obj.__module__},
            }
        elif isinstance(obj, pd.api.extensions.ExtensionDtype):
            return {"_type": "dtype", "data": str(obj)}
        elif isinstance(obj, slice):
            return {"_type": "slice", "data": (obj.start, obj.stop, obj.step)}
        elif isinstance(obj, range):
            return {
                "_type": "range",
                "data": {
                    '"start"': obj.start,
                    '"stop"': obj.stop,
                    '"step"': obj.step,
                },
            }
        return super().default(obj)

    def encode_obj(self, obj: Any) -> Any:
        if isinstance(obj, tuple):
            return {
                "_type": "tuple",
                "data": [self.encode_obj(v) for v in obj],
            }
        elif isinstance(obj, list):
            return [self.encode_obj(v) for v in obj]
        elif isinstance(obj, dict):
            return {self.encode(k): self.encode_obj(v) for k, v in obj.items()}
        return obj

    def encode(self, obj: Any) -> str:
        obj_transformed = self.encode_obj(obj)
        return super().encode(obj_transformed)

    @classmethod
    def encode_bytes(cls, obj: Any) -> bytes:
        encoder = cls()
        return (encoder.encode(obj)).encode("utf-8")


class SarusJSONDecoder(json.JSONDecoder):
    def decode(self, s: str, *args: Any, **kwargs: Any) -> Any:
        obj = super().decode(s, *args, **kwargs)
        return self.decode_obj(obj)

    def decode_obj(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            if "_type" in obj:
                data = self.decode_obj(obj["data"])
                if obj["_type"] == "tuple":
                    return tuple(self.decode_obj(v) for v in data)
                elif obj["_type"] == "numpy.ndarray":
                    return np.array(data)
                elif obj["_type"] == "pandas.DataFrame":
                    return pd.read_json(data, convert_dates=True)
                elif obj["_type"] == "pandas.Series":
                    return pd.read_json(data, typ="series", convert_dates=True)
                elif obj["_type"] == "pandas.Timestamp":
                    return pd.Timestamp(data)
                elif obj["_type"] == "datetime":
                    return parse(data)
                elif obj["_type"] == "timedelta":
                    return timedelta(seconds=data)
                elif obj["_type"] == "timezone":
                    return timezone(timedelta(seconds=data))
                elif obj["_type"] == "time":
                    return parse(data).time()
                elif obj["_type"] == "date":
                    return parse(data).date()
                elif obj["_type"] == "numpy.generic":
                    return np.array(data).item()
                elif obj["_type"] == "numpy.complex":
                    return np.complex128(complex(data["real"], data["imag"]))
                elif obj["_type"] == "pandas.Index":
                    cls = getattr(pd, obj["class"])
                    return cls(data, dtype=obj["dtype"])
                elif obj["_type"] == "pandas.MultiIndex":
                    return pd.MultiIndex(
                        levels=[pd.Index(level) for level in obj["levels"]],
                        codes=self.decode_obj(obj["codes"]),
                        names=obj["names"],
                    )
                elif obj["_type"] == "pandas.Period":
                    return pd.Period(data, freq=obj["freq"])
                elif obj["_type"] == "pandas.Timedelta":
                    return pd.to_timedelta(data)
                elif obj["_type"] == "pandas.Interval":
                    return pd.Interval(*data)
                elif obj["_type"] == "pandas.Categorical":
                    return pd.Categorical(
                        data,
                        categories=obj["categories"],
                        ordered=obj["ordered"],
                    )
                elif obj["_type"] == "class":
                    if data["module"] in ["builtins", "numpy", "pandas"]:
                        cls = getattr(
                            sys.modules[data["module"]], data["name"]
                        )
                        if isinstance(cls, type):
                            return cls
                        else:
                            raise ValueError("Decoded object is not a type")
                    else:
                        raise ValueError("Invalid module name")
                elif obj["_type"] == "dtype":
                    return pd.api.types.pandas_dtype(data)
                elif obj["_type"] == "slice":
                    return slice(*data)
                elif obj["_type"] == "range":
                    return range(data["start"], data["stop"], data["step"])
            return {self.decode(k): self.decode_obj(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.decode_obj(v) for v in obj]
        return obj

    @classmethod
    def decode_bytes(cls, b: bytes, *args: Any, **kwargs: Any) -> Any:
        decoder = cls()
        return decoder.decode(b.decode("utf-8"), *args, **kwargs)



from __future__ import annotations

import typing as t

from sarus_data_spec.dataspec_validator.signature import SarusBoundSignature
from sarus_data_spec.dataspec_validator.typing import PUPKind
import sarus_data_spec.typing as st

NO_TRANSFORM_ID = "no_transform_id"

try:
    from sarus_data_spec.sarus_query_builder.core.core import (
        OptimizableQueryBuilder,
    )
except ModuleNotFoundError as e_pandas_dp:
    if "sarus" not in str(e_pandas_dp):
        raise


@t.runtime_checkable
class ExternalOpImplementation(t.Protocol):
    """External Op implementation class.

    The `allowed_pup_args` is a list of combinations of arguments' names which
    are managed by the Op. The result of the Op will be PUP only if the set of
    PUP arguments passed to the Op are in this list.

    For instance, if we have an op that takes 3 arguments `a`, `b` and `c` and
    the `allowed_pup_args` are [{'a'}, {'b'}, {'a','b'}] then the following
    combinations will yield a PUP output:
        - `a` is a PUP dataspec, `b` and `c` are either not dataspecs or public
          dataspecs
        - `b` is a PUP dataspec, `a` and `c` are either not dataspecs or public
          dataspecs
        - `a` and `b` are PUP dataspecs, `c` is either not a dataspec or a
          public dataspec
    """

    def transform_id(self) -> str: ...

    def dp_equivalent_id(self) -> t.Optional[str]: ...

    def dp_equivalent(self) -> t.Optional[ExternalOpImplementation]: ...

    async def call(self, bound_signature: SarusBoundSignature) -> t.Any:
        """Compute the external op output value.

        DP implementation take additional arguments:
            - `seed` an integer used to parametrize rangom number generators
            - `budget` the privacy budget that can be spend in the op
            - `pe` the protected entity information of each row
        """

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        """Return the PUP properties of the transform.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, it is not PUP if we are aggregating
        the rows (axis=0) and it is PUP if we are aggregating the columns
        (axis=1).
        """

    def is_dp(self, bound_signature: SarusBoundSignature) -> bool:
        """Return True if the DP transform is compatible with the arguments.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, if we are aggregating the rows
        (axis=0), then there might be an equivalent DP transform but if we are
        aggregating the columns there might not (axis=1).
        """

    async def private_queries(
        self, signature: SarusBoundSignature
    ) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder: ...



import random
import typing as t

import pyarrow as pa

from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.manager.async_utils import async_iter
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.marginals import marginals as marg_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st


class ShuffleStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            privacy_unit_tracking_paths=None,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Shuffle(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a user_settings transform
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        arrow_batches = [
            batch async for batch in await self.parent_to_arrow(batch_size=1)
        ]
        random.shuffle(arrow_batches)
        return async_iter(
            pa.Table.from_batches(arrow_batches).to_batches(
                max_chunksize=batch_size
            )
        )

    async def size(self) -> st.Size:
        sizes = await self.parent_size()
        return size_builder(
            dataset=self.dataset, statistics=sizes.statistics()
        )

    async def multiplicity(self) -> st.Multiplicity:
        multiplicities = await self.parent_multiplicity()
        return multiplicity_builder(
            dataset=self.dataset, statistics=multiplicities.statistics()
        )

    async def bounds(self) -> st.Bounds:
        bounds = await self.parent_bounds()
        return bounds_builder(
            dataset=self.dataset, statistics=bounds.statistics()
        )

    async def marginals(self) -> st.Marginals:
        marginals = await self.parent_marginals()
        return marg_builder(
            dataset=self.dataset, statistics=marginals.statistics()
        )



from typing import AsyncIterator, List, Union
import typing as t
import warnings

import numpy as np
import pyarrow as pa
import gc

from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import DATA, DATASET_SLUGNAME
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.async_utils import async_iter
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.marginals import marginals as marginals_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.path import straight_path
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn("sqlalchemy not installed. No sampling on bigdata")

try:
    from sarus_data_spec.sarus_statistics.tasks.size.sample_visitor import (
        sampled_size,
    )
except ModuleNotFoundError as exception:
    # for the public repo
    if "sarus_data_spec.sarus_statistics" in str(exception.name):
        pass
    else:
        raise exception

try:
    from sarus_data_spec.manager.ops.sql_utils.bigdata import (
        path_to_quoted_string,
        sqlalchemy_query_to_string,
    )
    from sarus_data_spec.manager.ops.sql_utils.queries import nest_queries
    from sarus_data_spec.manager.ops.sql_utils.schema_translations import (
        async_sa_metadata_from_dataset,
    )
except ModuleNotFoundError:
    warnings.warn("sql utils not installed.")


class SampleStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            privacy_unit_tracking_paths=parent_schema.protobuf().privacy_unit,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Sample(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a sample transform
    """

    async def size(self) -> st.Size:
        parent_size = await self.parent_size()
        size_ratio = new_sampling_ratio(self.dataset, parent_size.statistics())
        if size_ratio >= 1:
            return parent_size
        return size_builder(
            self.dataset, sampled_size(parent_size.statistics(), size_ratio)
        )

    async def multiplicity(self) -> st.Multiplicity:
        parent_multiplicity = await self.parent_multiplicity()
        return multiplicity_builder(
            self.dataset, parent_multiplicity.statistics()
        )

    async def bounds(self) -> st.Bounds:
        parent_bounds = await self.parent_bounds()
        size_ratio = new_sampling_ratio(
            self.dataset, parent_bounds.statistics()
        )
        if size_ratio >= 1:
            return parent_bounds
        return bounds_builder(
            self.dataset, sampled_size(parent_bounds.statistics(), size_ratio)
        )

    async def marginals(self) -> st.Marginals:
        parent_marginals = await self.parent_marginals()
        size_ratio = new_sampling_ratio(
            self.dataset, parent_marginals.statistics()
        )
        if size_ratio >= 1:
            return parent_marginals
        return marginals_builder(
            self.dataset,
            sampled_size(parent_marginals.statistics(), size_ratio),
        )

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = t.cast(Dataset, self.parent())
        if parent.manager().is_big_data(parent):
            queries = await self.sql_implementation()
            assert queries is not None
            # rewrite queries as nested dict rather than dict[tuple,str]
            nested_queries = nest_queries(queries)
            schema = await self.dataset.async_schema()
            return await self.dataset.manager().async_sql(
                parent,
                query=nested_queries,
                batch_size=batch_size,
                result_type=schema.type(),
            )
        return await self._arrow_to_arrow(batch_size)

    async def _arrow_to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await sample_arrow_to_arrow(self, batch_size)

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        schema = await self.dataset.async_schema()
        parent_ds = t.cast(Dataset, self.parent())
        sample_spec = getattr(
            self.dataset.transform().protobuf().spec,
            self.dataset.transform().protobuf().spec.WhichOneof("spec"),  # type: ignore[arg-type]
        )

        previous_size = await parent_ds.manager().async_size(parent_ds)
        assert previous_size
        previous_stats = previous_size.statistics()

        sqlalchemy_metadata = await async_sa_metadata_from_dataset(
            self.dataset
        )

        queries = {}
        for table_path in schema.tables():
            curr_path = table_path.to_strings_list()[0]
            full_tablename = (
                curr_path[1:] if curr_path[0] == DATA else curr_path
            )
            sa_table_name = path_to_quoted_string(
                straight_path(full_tablename)
            )
            sa_table = sqlalchemy_metadata.tables[sa_table_name]
            if sample_spec.HasField("fraction"):
                fraction = sample_spec.fraction
                sample_query = sa.select(sa_table).where(
                    sa.func.random() < fraction
                )
            else:
                sample_query = (
                    sa.select(sa_table)
                    .order_by(sa.func.random())
                    .limit(
                        int(
                            sample_spec.size
                            * previous_stats.nodes_statistics(table_path)[
                                0
                            ].size()
                            / previous_stats.size()
                        )
                    )
                )

            queries[tuple(full_tablename)] = sqlalchemy_query_to_string(
                sample_query
            )
        return queries


def fast_gather(
    indices: Union[List[int], np.ndarray],
    batches: List[pa.RecordBatch],
    batch_size: int,
) -> AsyncIterator[pa.RecordBatch]:
    """
    A simple approach based on PyArrow's Table.take().
    """
    assert len(indices), "Indices must be non-empty"
    table = pa.Table.from_batches(batches, schema=batches[0].schema)
    return async_iter(table.take(indices).to_batches(max_chunksize=batch_size))


def new_sampling_ratio(
    dataset: st.Dataset, statistics: st.Statistics
) -> float:
    """From a transformed dataset which last transform is sample or
    differentiated_sample, get sampling ratio"""
    spec = dataset.transform().protobuf().spec
    transform_type = spec.WhichOneof("spec")
    assert transform_type
    sampling_spec = getattr(spec, transform_type)
    if sampling_spec.HasField("fraction"):
        return t.cast(float, sampling_spec.fraction)
    return t.cast(float, sampling_spec.size / statistics.size())


async def sample_arrow_to_arrow(
    op: StandardDatasetImplementation, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:
    spec = op.dataset.transform().protobuf().spec
    transform_type = spec.WhichOneof("spec")
    assert transform_type
    sampling_spec = getattr(spec, transform_type)
    seed = Scalar(sampling_spec.seed).value()
    generator = np.random.default_rng(seed)
    parent_batches = [batch async for batch in await op.parent_to_arrow()]
    parent_table = pa.Table.from_batches(parent_batches)

    if sampling_spec.HasField("fraction"):
        new_size = int(sampling_spec.fraction * parent_table.num_rows)
    else:
        new_size = sampling_spec.size

    parent_size = await op.parent_size()
    if new_size >= parent_size.statistics().size():
        del parent_batches, parent_table
        gc.collect()
        return await op.parent_to_arrow(batch_size)

    indices = generator.choice(
        parent_table.num_rows,
        replace=False,
        size=new_size,
    )
    return fast_gather(
        indices=indices,
        batches=parent_table.to_batches(max_chunksize=1000),
        batch_size=batch_size,
    )



from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

try:
    from xgboost import XGBClassifier, XGBRegressor
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC

# ------ CONSTRUCTORS -------


class xgb_classifier(ExternalOpImplementation):
    _transform_id = "xgboost.XGB_CLASSIFIER"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_leaves",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_bin",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="grow_policy",
            annotation=Optional[Literal[0, 1]],
            default=None,
        ),
        SarusParameter(
            name="learning_rate",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="verbosity",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="objective",
            annotation=Optional[Union[str, Callable]],
            default=None,
        ),
        SarusParameter(
            name="booster",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="tree_method",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="gamma",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="min_child_weight",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_delta_step",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="subsample",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bytree",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bylevel",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bynode",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="reg_alpha",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="reg_lambda",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="scale_pos_weight",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="base_score",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[np.random.RandomState, int]],
            default=None,
        ),
        SarusParameter(
            name="missing",
            annotation=float,
            default=np.nan,
        ),
        SarusParameter(
            name="num_parallel_tree",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="monotone_constraints",
            annotation=Optional[Union[Dict[str, int], str]],
            default=None,
        ),
        SarusParameter(
            name="interaction_constraints",
            annotation=Optional[Union[str, List[Tuple[str]]]],
            default=None,
        ),
        SarusParameter(
            name="importance_type",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="gpu_id",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="validate_parameters",
            annotation=Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="predictor",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="enable_categorical",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="max_cat_to_onehot",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_cat_threshold",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="eval_metric",
            annotation=Optional[Union[str, List[str], Callable]],
            default=None,
        ),
        SarusParameter(
            name="early_stopping_rounds",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="callbacks",
            annotation=Optional[List[Callable]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="kwargs",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="use_label_encoder",
            annotation=Optional[bool],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return XGBClassifier(**kwargs)


class xgb_regressor(ExternalOpImplementation):
    _transform_id = "xgboost.XGB_REGRESSOR"
    _signature = SarusSignature(
        SarusParameter(
            name="n_estimators",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="max_depth",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_leaves",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_bin",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="grow_policy",
            annotation=Optional[Literal[0, 1]],
            default=None,
        ),
        SarusParameter(
            name="learning_rate",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="verbosity",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="objective",
            annotation=Optional[
                Union[
                    str,
                    Callable[
                        [np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]
                    ],
                    None,
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="booster",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="tree_method",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="gamma",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="min_child_weight",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="max_delta_step",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="subsample",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bytree",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bylevel",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="colsample_bynode",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="reg_alpha",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="reg_lambda",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="scale_pos_weight",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="base_score",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[np.random.RandomState, int]],
            default=None,
        ),
        SarusParameter(
            name="missing",
            annotation=float,
            default=np.nan,
        ),
        SarusParameter(
            name="num_parallel_tree",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="monotone_constraints",
            annotation=Optional[Union[Dict[str, int], str]],
            default=None,
        ),
        SarusParameter(
            name="interaction_constraints",
            annotation=Optional[Union[str, List[Tuple[str]]]],
            default=None,
        ),
        SarusParameter(
            name="importance_type",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="gpu_id",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="validate_parameters",
            annotation=Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="predictor",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="enable_categorical",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="max_cat_to_onehot",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="max_cat_threshold",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="eval_metric",
            annotation=Optional[Union[str, List[str], Callable]],
            default=None,
        ),
        SarusParameter(
            name="early_stopping_rounds",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="callbacks",
            annotation=Optional[List[Callable]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="kwargs",
            annotation=Optional[Dict],
            default=None,
        ),
        SarusParameter(
            name="use_label_encoder",
            annotation=Optional[bool],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return XGBRegressor(**kwargs)


# ------ METHODS ------


class xgb_fit(ExternalOpImplementation):
    _transform_id = "xgboost.XGB_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Any,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="X",
            annotation=Union[pd.DataFrame, np.ndarray],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y",
            annotation=Union[pd.Series, np.ndarray],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="base_margin",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="eval_set",
            annotation=Optional[List[Tuple[np.ndarray, np.ndarray]]],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=Union[bool, int, None],
            default=True,
        ),
        SarusParameter(
            name="xgb_model",
            annotation=Any,
            default=None,
        ),
        SarusParameter(
            name="sample_weight_eval_set",
            annotation=Optional[List[np.ndarray]],
            default=None,
        ),
        SarusParameter(
            name="base_margin_eval_set",
            annotation=Optional[List[np.ndarray]],
            default=None,
        ),
        SarusParameter(
            name="feature_weights",
            annotation=Optional[np.ndarray],
            default=None,
        ),
        SarusParameter(
            name="callbacks",
            annotation=Optional[List[Callable]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit(**kwargs)



import typing as t

import pyarrow as pa

from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.schema import schema
import sarus_data_spec.typing as st


class ExtractStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            privacy_unit_tracking_paths=None,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Extract(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    an extract transform. It is used to generate the mocks.
    For the moment we extrac the full dataset so the size parameter
    is ignored. Also the seed which be found scalar parent is ignored.
    TODO: take into account size
    TODO: randomise the extract
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        return await self.parent_to_arrow(batch_size=batch_size)

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        """pass query to parents"""
        return None



from __future__ import annotations

import hashlib
import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.admin_utils import (
    compute_admin_data,
    create_admin_columns,
    merge_data_and_admin,
)
from sarus_data_spec.arrow.array import fit_array_to_schema_type_async_gen
from sarus_data_spec.arrow.conversion import to_pyarrow_table
from sarus_data_spec.arrow.schema import type_from_arrow_schema
from sarus_data_spec.constants import MAX_MAX_MULT, MULTIPLICITY
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.dataspec_validator.signature import (
    SarusBoundSignature,
    SarusSignature,
    SarusSignatureValue,
)
from sarus_data_spec.dataspec_validator.typing import PUPKind
from sarus_data_spec.manager.async_utils import async_iter
from sarus_data_spec.manager.ops.base import (
    DatasetImplementation,
    DatasetStaticChecker,
    DataspecStaticChecker,
    ScalarImplementation,
)
from sarus_data_spec.schema import schema as schema_builder
from sarus_data_spec.size import size as size_builder
from sarus_data_spec.transform import external, transform_id
import sarus_data_spec.protobuf as sp
import sarus_data_spec.statistics as sds
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

from .typing import NO_TRANSFORM_ID
from .utils import static_arguments

try:
    from sarus_data_spec.sarus_query_builder.core.core import (
        OptimizableQueryBuilder,
    )
except ModuleNotFoundError as e_pandas_dp:
    if "sarus" not in str(e_pandas_dp):
        raise

DEFAULT_MAX_MAX_MULT = "1000.0"


class ExternalScalarStaticChecker(DataspecStaticChecker):
    async def private_queries(self) -> t.List[st.PrivateQuery]:
        """Return the PrivateQueries summarizing DP characteristics."""
        implementation = external_implementation(self.dataspec.transform())
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        return await implementation.private_queries(bound_signature)

    def is_dp(self) -> bool:
        """Checks if the transform is DP and compatible with the arguments."""
        implementation = external_implementation(self.dataspec.transform())
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        return implementation.is_dp(bound_signature)

    def is_dp_able(self) -> bool:
        """Checks if the dataspec has a transform that either has a DP
        equivalent, allowing the rewritten dataspec to be considered DP
        if the input rewritten PUP token is not None."""
        if self.is_dp():
            return True

        transform = self.dataspec.transform()
        implementation = external_implementation(transform)
        dp_implementation = implementation.dp_equivalent()
        if dp_implementation is None:
            return False
        else:
            bound_signature = implementation.signature().bind_dataspec(
                self.dataspec
            )
            return dp_implementation.is_dp(bound_signature)

    def is_dp_writable(self, public_context: t.Collection[str]) -> bool:
        """Statically check if a DP transform is applicable in this position.

        This verification is common to all dataspecs and is true if:
            - the dataspec is transformed and its transform has an equivalent
            DP transform
            - the DP transform's required PUP arguments are PUP and aligned
            (i.e. same PUP token)
            - other dataspecs arguments are public
        """
        transform = self.dataspec.transform()
        implementation = external_implementation(transform)
        dp_implementation = implementation.dp_equivalent()
        if dp_implementation is None:
            return False
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        if not dp_implementation.is_dp(bound_signature):
            return False

        return bound_signature.rewritten_pup_token() is not None

    def dp_transform(self) -> t.Optional[st.Transform]:
        """Return the dataspec's DP equivalent transform if existing."""
        transform = self.dataspec.transform()
        op_implementation = external_implementation(transform)
        py_args, py_kwargs, ds_args_pos, ds_types = static_arguments(transform)

        dp_implementation = op_implementation.dp_equivalent()
        if dp_implementation is None:
            return None

        dp_transform_id = dp_implementation.transform_id()
        assert dp_transform_id is not None

        # Types won't be used since budget & seed are scalars
        ds_types["budget"] = ""
        ds_types["seed"] = ""

        return external(
            dp_transform_id,
            py_args=py_args,
            py_kwargs=py_kwargs,
            ds_args_pos=ds_args_pos,
            ds_types=ds_types,
        )

    def pup_transform(self) -> t.Optional[st.Transform]:
        transform = self.dataspec.transform()
        pup_implementation = pup_external_implementation(self.dataspec)
        if pup_implementation is None:
            return None
        py_args, py_kwargs, ds_args_pos, ds_types = static_arguments(transform)
        dp_transform_id = pup_implementation.transform_id()
        assert dp_transform_id is not None
        return external(
            dp_transform_id,
            py_args=py_args,
            py_kwargs=py_kwargs,
            ds_args_pos=ds_args_pos,
            ds_types=ds_types,
        )

    async def query_builder(self) -> OptimizableQueryBuilder:
        transform = self.dataspec.transform()
        implementation = external_implementation(transform)

        dp_implementation = (
            implementation
            if implementation.is_dp_equivalent()
            else implementation.dp_equivalent()
        )
        if dp_implementation is None:
            raise ValueError(
                """The dataspec does not have a dp equivalent,
                thus it does not have a query builder"""
            )
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )
        return await dp_implementation.query_builder(bound_signature)


class ExternalDatasetStaticChecker(
    ExternalScalarStaticChecker, DatasetStaticChecker
):
    def __init__(self, dataset: st.Dataset):
        super().__init__(dataset)
        self.dataset = dataset

    def pup_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        """Return the dataspec's PUP token."""
        transform = self.dataspec.transform()
        implementation = external_implementation(transform)
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )

        input_token = bound_signature.pup_token()
        if input_token is None:
            return None

        pup_kind = implementation.pup_kind(bound_signature)
        if pup_kind == PUPKind.NOT_PUP:
            return None
        elif pup_kind == PUPKind.TOKEN_PRESERVING:
            return input_token
        else:  # PUP or ROW
            h = hashlib.md5(usedforsecurity=False)
            h.update(input_token.encode("ascii"))
            h.update(transform.protobuf().SerializeToString())
            new_token = h.hexdigest()
            return new_token

    def rewritten_pup_token(
        self, public_context: t.Collection[str]
    ) -> t.Optional[str]:
        transform = self.dataspec.transform()
        implementation = pup_external_implementation(self.dataspec)
        if implementation is None:
            return None
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )

        input_token = bound_signature.rewritten_pup_token()
        if input_token is None:
            return None

        pup_kind = implementation.pup_kind(bound_signature)
        if pup_kind == PUPKind.NOT_PUP:
            return None
        elif pup_kind == PUPKind.TOKEN_PRESERVING:
            return input_token
        else:  # PUP or ROW
            h = hashlib.md5(usedforsecurity=False)
            h.update(input_token.encode("ascii"))
            h.update(transform.protobuf().SerializeToString())
            new_token = h.hexdigest()
            return new_token

    def is_pup_able(self) -> bool:
        """Checks if the dataspec has a transform that either has a PUP
        equivalent or does not require one, allowing the rewritten dataspec to
        be considered 'PUP' if the input rewritten PUP token is not None."""
        implementation = pup_external_implementation(self.dataspec)
        if implementation is None:
            return False
        # probably not needed
        bound_signature = implementation.signature().bind_dataspec(
            self.dataspec
        )

        pup_kind = implementation.pup_kind(bound_signature)
        return not pup_kind == PUPKind.NOT_PUP

    async def schema(self) -> st.Schema:
        """Computes the schema of the dataspec.

        The schema is computed by computing the synthetic data value and
        converting the Pyarrow schema to a Sarus schema.q
        """
        syn_variant = self.dataset.variant(kind=st.ConstraintKind.SYNTHETIC)
        assert syn_variant is not None
        assert syn_variant.prototype() == sp.Dataset

        syn_dataset = t.cast(st.Dataset, syn_variant)
        arrow_iterator = await compute_external_to_arrow(
            syn_dataset, batch_size=1000
        )
        first_batch = await arrow_iterator.__anext__()
        schema = first_batch.schema

        schema_type = type_from_arrow_schema(schema)

        # retrieve max_mul from parent
        parents_args, parents_kwargs = self.dataset.parents()
        parents_args.extend(parents_kwargs.values())
        parents_args = [el for el in parents_args if isinstance(el, Dataset)]
        if len(parents_args) > 1 or len(parents_args) == 0:
            max_mult = ""  # we cannot infer max_mult in this case
            max_max_mult = DEFAULT_MAX_MAX_MULT
        else:
            parent_ds = t.cast(st.Dataset, parents_args[0])
            parent_schema = await self.dataset.manager().async_schema(
                parent_ds
            )
            max_mult = await retrieve_max_mult_info_from_parent(
                parent_ds=parent_ds, parent_schema=parent_schema
            )
            max_max_mult = parent_schema.properties().get(
                MAX_MAX_MULT, DEFAULT_MAX_MAX_MULT
            )
        properties = {}
        if self.dataset.is_pup():
            # If the dataset is PUP then the schema of the real data should
            # have protection but the synthetic data might not have it. We
            # need to add it manually.
            schema_type = sdt.protected_type(schema_type)
            # now take care of multiplicity
            implementation = external_implementation(self.dataset.transform())
            bound_signature = implementation.signature().bind_dataspec(
                self.dataset
            )
            pup_kind = implementation.pup_kind(bound_signature)
            if pup_kind == PUPKind.TOKEN_PRESERVING:
                if max_mult != "":
                    properties[MULTIPLICITY] = max_mult
                    properties[MAX_MAX_MULT] = max_mult
                else:
                    properties[MAX_MAX_MULT] = max_max_mult

            else:
                # in this case, rows are not preserved,
                # so the max multiplicity might change,
                # we take as bound the older one, this is
                # an approximation and may cause bias
                # if the dataset rows increase
                properties = {
                    MAX_MAX_MULT: max_mult if max_mult != "" else max_max_mult
                }
        else:
            properties = {}
        return schema_builder(
            self.dataset, schema_type=schema_type, properties=properties
        )


class ExternalDatasetOp(DatasetImplementation):
    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        batches_async_it = await compute_external_to_arrow(
            self.dataset, batch_size
        )
        schema = await self.dataset.async_schema()
        # make sure the arrow array is conform to its schema.
        return fit_array_to_schema_type_async_gen(
            batches_async_it, schema.type()
        )

    async def size(self) -> st.Size:
        implementation = external_implementation(self.dataset.transform())
        bound_signature = implementation.signature().bind_dataspec(
            self.dataset
        )
        pup_kind = implementation.pup_kind(bound_signature)
        if pup_kind == PUPKind.TOKEN_PRESERVING:
            parents_args, parents_kwargs = self.dataset.parents()
            parents_args.extend(parents_kwargs.values())
            parents_args = [
                el for el in parents_args if isinstance(el, Dataset)
            ]
            assert len(parents_args) == 1
            parent_ds = t.cast(st.Dataset, parents_args[0])
            sizes = await self.dataset.manager().async_size(parent_ds)
            return size_builder(
                dataset=self.dataset, statistics=sizes.statistics()
            )
        elif pup_kind == PUPKind.PUP:
            syn_variant = self.dataset.variant(
                kind=st.ConstraintKind.SYNTHETIC
            )
            assert syn_variant is not None
            syn_variant = t.cast(Dataset, syn_variant)
            import warnings

            warnings.warn(
                "Using hack to compute sizes of dataset with synthetic variant"
            )
            size = len(
                await syn_variant.manager().async_to_pandas(syn_variant)
            )
            return size_builder(
                dataset=self.dataset,
                statistics=sds.Struct(fields={}, multiplicity=1.0, size=size),
            )
        else:
            raise NotImplementedError(
                f"Size not implemented for transform"
                f" {self.dataset.transform().spec()}"
            )


class ExternalScalarOp(ScalarImplementation):
    async def value(self) -> t.Any:
        transform = self.scalar.transform()
        ds_args, ds_kwargs = self.scalar.parents()
        return await async_compute_external_value(
            transform, *ds_args, **ds_kwargs
        )


async def async_compute_external_value(
    transform: st.Transform,
    *ds_args: t.Union[st.DataSpec, st.Transform],
    **ds_kwargs: t.Union[st.DataSpec, st.Transform],
) -> t.Any:
    """Compute the value of an external transform applied on Dataspecs.

    This function computes the output value without manipulating the
    corresponding Dataspec. This is useful when we need to have access
    to the value of a Dataspec before its creation:
      - for computing a Mock value and inferring if the result is
        a Scalar or a Dataset.
    """
    implementation = external_implementation(transform)
    bound_signature = implementation.signature().bind(
        transform, *ds_args, **ds_kwargs
    )
    bound_signature.static_validation()
    data = await implementation.compute(bound_signature)
    return data


class ExternalOpImplementation:
    """External PUP op implementation class.

    This class wraps together several elements of an external op
    implementation:
        - `call` is the function that computes the output value from the
          input(s) value(s).
    """

    _transform_id: str = NO_TRANSFORM_ID
    _dp_equivalent_id: t.Optional[str] = None
    _non_dp_equivalent_id: t.Optional[str] = None
    _pup_equivalent_id: t.Optional[str] = None
    _non_pup_equivalent_id: t.Optional[str] = None
    _signature: t.Optional[SarusSignature] = None

    def transform_id(self) -> str:
        return self._transform_id

    def dp_equivalent_id(self) -> t.Optional[str]:
        return self._dp_equivalent_id

    def is_dp_equivalent(self) -> bool:
        return self._non_dp_equivalent_id is not None

    def dp_equivalent(self) -> t.Optional[ExternalOpImplementation]:
        if not self._dp_equivalent_id:
            return None

        return external_implementation_from_id(self._dp_equivalent_id)

    def pup_equivalent(self) -> t.Optional[ExternalOpImplementation]:
        if not self._pup_equivalent_id:
            return None

        return external_implementation_from_id(self._pup_equivalent_id)

    def signature(self) -> SarusSignature:
        if self._signature is not None:
            return self._signature

        if self._non_dp_equivalent_id is None:
            raise ValueError(
                f"External implementation {self.transform_id()} has no "
                "signature defined and no non-DP equivalent."
            )

        non_dp_signature = external_implementation_from_id(
            self._non_dp_equivalent_id
        ).signature()
        return non_dp_signature.make_dp()

    async def compute(self, bound_signature: SarusBoundSignature) -> t.Any:
        if self.is_dp(bound_signature):
            return await self.call_dp(bound_signature)
        else:
            signature_value = await bound_signature.collect_signature_value()
            return self.call(signature_value)

    def call(self, signature_value: SarusSignatureValue) -> t.Any:
        raise NotImplementedError

    async def call_dp(self, bound_signature: SarusBoundSignature) -> t.Any:
        """DP ops `call` need to be async to compute schema, tasks, etc"""
        raise NotImplementedError

    def pup_kind(self, bound_signature: SarusBoundSignature) -> PUPKind:
        """Return the PUP properties of the transform.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, it is not PUP if we are aggregating
        the rows (axis=0) and it is PUP if we are aggregating the columns
        (axis=1).
        """
        # Default implementation
        return PUPKind.NOT_PUP

    def is_dp(self, bound_signature: SarusBoundSignature) -> bool:
        """Return True if the DP transform is compatible with the arguments.

        It takes the transform arguments as input because it can depend on some
        transform parameters. For instance, if we are aggregating the rows
        (axis=0), then there might be an equivalent DP transform but if we are
        aggregating the columns there might not (axis=1).
        """
        # Default implementation
        return False

    async def private_queries(
        self, signature: SarusBoundSignature
    ) -> t.List[st.PrivateQuery]:
        """Takes as input the args of the transform (static and dynamic)."""
        if not signature.is_dp():
            return []
        queries, _ = await self.private_queries_and_task(signature)
        return queries

    async def query_builder(
        self, signature: SarusBoundSignature
    ) -> OptimizableQueryBuilder:
        raise NotImplementedError

    async def private_queries_and_task(
        self, signature: SarusBoundSignature
    ) -> t.Tuple[t.List[st.PrivateQuery], st.Task]:
        raise NotImplementedError

    def callable(
        self, composed_transform: st.Transform
    ) -> t.Callable[..., t.Awaitable[t.Any]]:
        """Build the transform's async callable.

        The function takes an undefined number of named arguments.

        It first collects the current transform's signature concrete values
        using the passed variables' values. The concrete values are stored in a
        SarusBoundSignature object so we can compute the current transform's
        output by simply using the implementation's `call` method.
        """
        lambda_signature = self.signature().bind_composed(composed_transform)
        previous_callable = lambda_signature.callable()

        def composed_callable(*vars: t.Any, **kwvars: t.Any) -> t.Any:
            signature_value = previous_callable(*vars, **kwvars)
            return self.call(signature_value)

        return composed_callable


def external_implementation(
    transform: st.Transform,
) -> ExternalOpImplementation:
    """Return the implementation of an external op from a DataSpec.

    The mapping is done by the config file.
    """
    assert transform and transform.is_external()
    id = transform_id(transform)
    return external_implementation_from_id(id)


def pup_external_implementation(
    dataspec: st.DataSpec,
) -> t.Optional[ExternalOpImplementation]:
    transform = dataspec.transform()
    implementation = external_implementation(transform)
    bound_signature = implementation.signature().bind_dataspec(dataspec)
    pup_kind = implementation.pup_kind(bound_signature)
    if not pup_kind == PUPKind.NOT_PUP:
        return implementation
    else:
        pup_implementation = implementation.pup_equivalent()
        if pup_implementation is None:
            return None
        else:
            pup_kind = pup_implementation.pup_kind(bound_signature)
            if pup_kind == PUPKind.NOT_PUP:
                return None
            else:
                return pup_implementation


def external_implementation_from_id(id: str) -> ExternalOpImplementation:
    # Imported here to avoid circular imports
    from . import ROUTING

    library, op_name = id.split(".")
    op_implementation = t.cast(
        ExternalOpImplementation, ROUTING[library][op_name]
    )
    return op_implementation


async def retrieve_max_mult_info_from_parent(
    parent_ds: st.Dataset, parent_schema: st.Schema
) -> str:
    """If the dataset is transformed by an external the max_mult
    lies in the schema while for standard transforms we can derive
     the multiplicity from the parent directly"""
    if not parent_ds.is_transformed():
        return ""
    if parent_ds.transform().is_external():
        max_mult = parent_schema.properties().get(MULTIPLICITY, "")
    else:
        try:
            max_mult = str(
                (await parent_ds.manager().async_multiplicity(parent_ds))
                .statistics()
                .multiplicity()
            )
        except Exception:
            # this is for the mock, not very nice
            max_mult = ""
    return max_mult


async def compute_external_to_arrow(
    dataset: st.Dataset, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:
    """It computes arrow record batches of a dataset from the external op implementation"""
    transform = dataset.transform()
    implementation = external_implementation(transform)
    bound_signature = implementation.signature().bind_dataspec(dataset)
    bound_signature.static_validation()

    if dataset.is_pup():
        result = await implementation.compute(bound_signature)
        if (
            isinstance(result, pd.Series)
            and implementation.pup_kind(bound_signature) == PUPKind.ROW
        ):
            # Reformat the series as a dataframe with a single row
            result = result.to_frame().transpose()
        ds_result = t.cast(st.DatasetCastable, result)
        admin_data = await bound_signature.admin_data()
        output_admin_data = compute_admin_data(admin_data, ds_result)
        data_table = to_pyarrow_table(ds_result)
        table = merge_data_and_admin(data_table, output_admin_data)

    else:
        result = await implementation.compute(bound_signature)
        data_table = to_pyarrow_table(result)
        table = create_admin_columns(data_table)

    return async_iter(table.to_batches(max_chunksize=batch_size))



from typing import Any, List, Literal, Optional, Union

import numpy.typing as npt

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn import metrics
except ModuleNotFoundError:
    spmatrix = Any


# metrics
class sk_accuracy_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ACCURACY_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="normalize",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.accuracy_score(**kwargs)


class sk_average_precision_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_AVERAGE_PRECISION_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_score",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="average",
            annotation=Optional[
                Literal["micro", "samples", "weighted", "macro"]
            ],
            default="macro",
        ),
        SarusParameter(
            name="pos_label",
            annotation=Union[int, str],
            default=1,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.average_precision_score(**kwargs)


class sk_classification_report(ExternalOpImplementation):
    _transform_id = "sklearn.SK_CLASSIFICATION_REPORT"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="target_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="digits",
            annotation=int,
            default=2,
        ),
        SarusParameter(
            name="output_dict",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="zero_division",
            annotation=Literal["warn", 0, 1],
            default="warn",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.classification_report(**kwargs)


class sk_confusion_matrix(ExternalOpImplementation):
    _transform_id = "sklearn.SK_CONFUSION_MATRIX"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="normalize",
            annotation=Optional[Literal["true", "pred", "all"]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.confusion_matrix(**kwargs)


class sk_f1_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_F1_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="pos_label",
            annotation=Union[int, str],
            default=1,
        ),
        SarusParameter(
            name="average",
            annotation=Optional[
                Literal["micro", "macro", "samples", "weighted", "binary"]
            ],
            default="binary",
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="zero_division",
            annotation=Literal["warn", 0, 1],
            default="warn",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.f1_score(**kwargs)


class sk_precision_recall_curve(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PRECISION_RECALL_CURVE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="probas_pred",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="pos_label",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.precision_recall_curve(**kwargs)


class sk_precision_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PRECISION_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="pos_label",
            annotation=Union[int, str],
            default=1,
        ),
        SarusParameter(
            name="average",
            annotation=Optional[
                Literal["micro", "macro", "samples", "weighted", "binary"]
            ],
            default="binary",
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="zero_division",
            annotation=Literal["warn", 0, 1],
            default="warn",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.precision_score(**kwargs)


class sk_recall_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_RECALL_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_pred",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="pos_label",
            annotation=Union[str, int],
            default=1,
        ),
        SarusParameter(
            name="average",
            annotation=Optional[
                Literal["micro", "macro", "samples", "weighted", "binary"]
            ],
            default="binary",
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="zero_division",
            annotation=Literal["warn", 0, 1],
            default="warn",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.recall_score(**kwargs)


class sk_roc_auc_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ROC_AUC_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_score",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="average",
            annotation=Optional[
                Literal["micro", "macro", "samples", "weighted"]
            ],
            default="macro",
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="max_fpr",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="multi_class",
            annotation=Literal["raise", "ovo", "ovr"],
            default="raise",
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.roc_auc_score(**kwargs)


class sk_auc(ExternalOpImplementation):
    _transform_id = "sklearn.SK_AUC"
    _signature = SarusSignature(
        SarusParameter(
            name="x",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.auc(**kwargs)


class sk_roc_curve(ExternalOpImplementation):
    _transform_id = "sklearn.SK_ROC_CURVE"
    _signature = SarusSignature(
        SarusParameter(
            name="y_true",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y_score",
            annotation=npt.ArrayLike,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="pos_label",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="drop_intermediate",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return metrics.roc_curve(**kwargs)



from __future__ import annotations

from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from .external_op import ExternalOpImplementation

try:
    from imblearn import over_sampling, pipeline, under_sampling
except ModuleNotFoundError:
    pass  # error message in typing.py

SamplingStrategy = Literal[
    "majority", "not minority", "not majority", "all", "auto"
]


# ------ CONSTRUCTORS ------
class imb_pipeline(ExternalOpImplementation):
    _transform_id = "imblearn.IMB_PIPELINE"
    _signature = SarusSignature(
        SarusParameter(
            name="steps",
            annotation=List[Tuple[str, Any]],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="memory",
            annotation=Optional,
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pipeline.Pipeline(**kwargs)


class imb_random_under_sampler(ExternalOpImplementation):
    _transform_id = "imblearn.IMB_RANDOM_UNDER_SAMPLER"
    _signature = SarusSignature(
        SarusParameter(
            name="sampling_strategy",
            annotation=Union[float, SamplingStrategy, Callable, Dict],
            default="auto",
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="replacement",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return under_sampling.RandomUnderSampler(**kwargs)


class imb_smotenc(ExternalOpImplementation):
    _transform_id = "imblearn.IMB_SMOTENC"
    _signature = SarusSignature(
        SarusParameter(
            name="categorical_features",
            annotation=Union[List[int], List[bool]],
        ),
        SarusParameter(
            name="sampling_strategy",
            annotation=Union[float, SamplingStrategy, Callable, Dict],
            default="auto",
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="k_neighbors",
            annotation=Union[int, object],
            default=5,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return over_sampling.SMOTENC(**kwargs)



from typing import Any, Dict, List, Optional, Sequence, Set, Union

from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
from sarus_data_spec.manager.async_utils import sync
from sarus_data_spec.manager.ops.processor import routing
import sarus_data_spec.typing as st

try:
    from sarus_data_spec.sarus_query_builder.builders.composed_builder import (
        ComposedBuilder,
    )
    from sarus_data_spec.sarus_query_builder.core.core import (
        OptimizableQueryBuilder,
    )
except ModuleNotFoundError as e_pandas_dp:
    if "sarus" not in str(e_pandas_dp):
        raise

    # Placeholder classes
    class ComposedBuilder:  # type: ignore
        pass

    class OptimizableQueryBuilder:  # type: ignore
        pass


class GraphBuilder(ComposedBuilder):
    """Graph builder"""

    def __init__(
        self,
        dataspec: st.DataSpec,
        builders: Sequence[OptimizableQueryBuilder],
        dp_applicable_dataspecs: Sequence[st.DataSpec],
        weights: Sequence[float],
    ):
        assert len(builders) == len(weights)
        assert len(builders) == len(dp_applicable_dataspecs)
        self._dataset = dataspec  # type: ignore
        self._builders = builders
        self._weights = weights
        self._dp_applicable_dataspecs = dp_applicable_dataspecs

    def split_graph_builder_in_dict_of_builders(
        self,
    ) -> Dict[str, Any]:
        dict_builder = {}
        for builder, dataspec in zip(
            self._builders, self._dp_applicable_dataspecs
        ):
            dict_builder[dataspec.uuid()] = builder
        return dict_builder


def simple_graph_builder(
    dataspec: st.DataSpec,
    builders: Sequence[OptimizableQueryBuilder],
    dp_applicable_dataspecs: Sequence[st.DataSpec],
) -> GraphBuilder:
    return GraphBuilder(
        dataspec, builders, dp_applicable_dataspecs, [1] * len(builders)
    )


def build_and_fit_graph_builder(
    dataspec: st.DataSpec, privacy_limit: st.PrivacyLimit
) -> GraphBuilder:
    delta_epsilon_dict = privacy_limit.delta_epsilon_dict()
    delta, epsilon = next(iter(delta_epsilon_dict.items()))
    # build query builder for the computation graph
    graph_query_builder = build_graph_query_builder(dataspec)
    if graph_query_builder is not None and epsilon > 0:
        graph_query_builder.fit([(epsilon, delta)], (0, epsilon))
    return graph_query_builder


def extract_privacy_limit(
    query_buidler: OptimizableQueryBuilder,
) -> Optional[st.PrivacyLimit]:
    if not query_buidler.is_fitted():
        return None
    else:
        epsilon_deltas_budget = query_buidler.epsilon_deltas_budget
        assert epsilon_deltas_budget is not None
        epsilon, delta = list(epsilon_deltas_budget)[0]
        return DeltaEpsilonLimit({delta: epsilon})


def build_graph_query_builder(dataspec: st.DataSpec) -> GraphBuilder:
    """Return all the private queries from a dataspec to the sources."""

    class BuilderVisitor(st.Visitor, st.TransformVisitor):
        visited: Set[Union[st.DataSpec, st.Transform]] = set()
        builders: List[ComposedBuilder] = list()
        dp_applicable_dataspecs: List[st.DataSpec] = list()

        def all(self, visited: Union[st.DataSpec, st.Transform]) -> None:
            pass

        def transformed(
            self,
            visited: st.DataSpec,
            transform: st.Transform,
            *arguments: st.DataSpec,
            **named_arguments: st.DataSpec,
        ) -> None:
            if visited not in self.visited:
                if visited.is_dp_writable():
                    _, StaticChecker = routing.get_op(visited)
                    builder = sync(StaticChecker(visited).query_builder())
                    self.builders.append(builder)
                    self.dp_applicable_dataspecs.append(visited)

                for argument in arguments:
                    argument.accept(self)
                for _, argument in named_arguments.items():
                    argument.accept(self)
                self.visited.add(visited)

        def other(self, visited: Union[st.DataSpec, st.Transform]) -> None:
            if visited not in self.visited:
                self.visited.add(visited)

        def composed(
            self,
            visited: st.Transform,
            transform: st.Transform,
            *arguments: st.Transform,
            **named_arguments: st.Transform,
        ) -> None:
            for argument in arguments:
                argument.accept(self)
            for _, argument in named_arguments.items():
                argument.accept(self)

        def variable(
            self,
            visited: st.Transform,
            name: str,
            position: int,
        ) -> None:
            pass

    visitor = BuilderVisitor()
    dataspec.accept(visitor)
    builders = visitor.builders
    dp_applicable_dataspecs = visitor.dp_applicable_dataspecs
    return simple_graph_builder(dataspec, builders, dp_applicable_dataspecs)



import typing as t

from sarus_data_spec.json_serialisation import SarusJSONDecoder
import sarus_data_spec.typing as st


def static_arguments(
    transform: st.Transform,
) -> t.Tuple[
    t.Dict[int, t.Any],
    t.Dict[str, t.Any],
    t.List[int],
    t.Dict[t.Union[str, int], str],
]:
    """Return the external arguments serialized in the protobuf."""
    assert transform and transform.is_external()

    transform_spec = transform.protobuf().spec
    external_args = SarusJSONDecoder.decode_bytes(
        transform_spec.external.named_arguments
    )
    py_args = external_args["py_args"]
    py_kwargs = external_args["py_kwargs"]
    ds_args_pos = external_args["ds_args_pos"]
    ds_types = external_args["ds_types"]
    return py_args, py_kwargs, ds_args_pos, ds_types


def static_and_dynamic_arguments(
    transform: st.Transform, *ds_args: st.DataSpec, **ds_kwargs: st.DataSpec
) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
    """Return all the external arguments.

    This returns static arguments interleaved with Dataspecs.
    """
    py_args, py_kwargs, ds_args_pos, _ = static_arguments(transform)
    pos_values = {pos: val for pos, val in zip(ds_args_pos, ds_args)}
    kwargs = {**py_kwargs, **ds_kwargs}
    pos_args = {**pos_values, **py_args}
    args = [pos_args[i] for i in range(len(pos_args))]

    return args, kwargs



from typing import Any, Dict, List, Optional, Union

import numpy.typing as npt

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from shap import Explanation
    import shap
except ModuleNotFoundError:
    spmatrix = Any
    Explanation = Any


# ------ CONSTRUCTORS -------
class shap_explanation(ExternalOpImplementation):
    _transform_id = "shap.SHAP_EXPLANATION"
    _signature = SarusSignature(
        SarusParameter(
            name="values",
            annotation=Union[npt.ArrayLike, spmatrix],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="base_values",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="data",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="display_data",
            annotation=Optional[Dict[str, npt.ArrayLike]],
            default=None,
        ),
        SarusParameter(
            name="instance_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="output_names",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="output_indexes",
            annotation=Optional[List[int]],
            default=None,
        ),
        SarusParameter(
            name="lower_bounds",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="upper_bounds",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="error_std",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="main_effects",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="hierarchical_values",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="clustering",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
        ),
        SarusParameter(
            name="compute_time",
            annotation=Optional[float],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.Explanation(**kwargs)


# ------ Explanation METHODS ------


class shap_values(ExternalOpImplementation):
    _transform_id = "shap.SHAP_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explanation,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.values


class shap_sum(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SUM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Explanation,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.sum()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/scalar.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%sarus_data_spec/protobuf/scalar.proto\x12\x0fsarus_data_spec\"\xaa\x08\n\x06Scalar\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12*\n\x04spec\x18\x04 \x01(\x0b\x32\x1c.sarus_data_spec.Scalar.Spec\x12;\n\nproperties\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Scalar.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xbd\x02\n\x04Spec\x12:\n\x0btransformed\x18\x01 \x01(\x0b\x32#.sarus_data_spec.Scalar.TransformedH\x00\x12.\n\x05model\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Scalar.ModelH\x00\x12\x43\n\x0eprivacy_params\x18\x03 \x01(\x0b\x32).sarus_data_spec.Scalar.PrivacyParametersH\x00\x12\x39\n\x0brandom_seed\x18\x04 \x01(\x0b\x32\".sarus_data_spec.Scalar.RandomSeedH\x00\x12\x41\n\x0fsynthetic_model\x18\x05 \x01(\x0b\x32&.sarus_data_spec.Scalar.SyntheticModelH\x00\x42\x06\n\x04spec\x1a\xbc\x01\n\x0bTransformed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12P\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\x37.sarus_data_spec.Scalar.Transformed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xad\x01\n\x05Model\x12I\n\x10pretrained_model\x18\x01 \x01(\x0b\x32-.sarus_data_spec.Scalar.Model.PretrainedModelH\x00\x1aQ\n\x0fPretrainedModel\x12\x1d\n\x15\x66oundation_model_name\x18\x01 \x01(\t\x12\x1f\n\x17peft_initialisation_uri\x18\x02 \x01(\tB\x06\n\x04type\x1a}\n\x11PrivacyParameters\x12?\n\x06points\x18\x01 \x03(\x0b\x32/.sarus_data_spec.Scalar.PrivacyParameters.Point\x1a\'\n\x05Point\x12\x0f\n\x07\x65psilon\x18\x01 \x01(\x01\x12\r\n\x05\x64\x65lta\x18\x02 \x01(\x01\x1a\x1b\n\nRandomSeed\x12\r\n\x05value\x18\x01 \x01(\x05\x1a\x10\n\x0eSyntheticModelb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.scalar_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SCALAR_PROPERTIESENTRY._options = None
  _SCALAR_PROPERTIESENTRY._serialized_options = b'8\001'
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._options = None
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _SCALAR._serialized_start=59
  _SCALAR._serialized_end=1125
  _SCALAR_PROPERTIESENTRY._serialized_start=215
  _SCALAR_PROPERTIESENTRY._serialized_end=264
  _SCALAR_SPEC._serialized_start=267
  _SCALAR_SPEC._serialized_end=584
  _SCALAR_TRANSFORMED._serialized_start=587
  _SCALAR_TRANSFORMED._serialized_end=775
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_start=722
  _SCALAR_TRANSFORMED_NAMEDARGUMENTSENTRY._serialized_end=775
  _SCALAR_MODEL._serialized_start=778
  _SCALAR_MODEL._serialized_end=951
  _SCALAR_MODEL_PRETRAINEDMODEL._serialized_start=862
  _SCALAR_MODEL_PRETRAINEDMODEL._serialized_end=943
  _SCALAR_PRIVACYPARAMETERS._serialized_start=953
  _SCALAR_PRIVACYPARAMETERS._serialized_end=1078
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_start=1039
  _SCALAR_PRIVACYPARAMETERS_POINT._serialized_end=1078
  _SCALAR_RANDOMSEED._serialized_start=1080
  _SCALAR_RANDOMSEED._serialized_end=1107
  _SCALAR_SYNTHETICMODEL._serialized_start=1109
  _SCALAR_SYNTHETICMODEL._serialized_end=1125
# @@protoc_insertion_point(module_scope)



import typing as t
import warnings

import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import DATA, DATASET_SLUGNAME, MULTIPLICITY
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.manager.ops.processor.standard.visitor_selector import (  # noqa : E501
    select_rows,
)
from sarus_data_spec.marginals import marginals as marg_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.path import Path, straight_path
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn("sqlalchemy not installed. No sampling on bigdata")

try:
    from sarus_data_spec.manager.ops.sql_utils.bigdata import (
        path_to_quoted_string,
        sqlalchemy_query_to_string,
    )
    from sarus_data_spec.manager.ops.sql_utils.schema_translations import (
        async_sa_metadata_from_dataset,
    )
except ModuleNotFoundError:
    warnings.warn("sql utils not installed.")


class GetItemStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        sub_types = parent_schema.data_type().sub_types(path)
        assert len(sub_types) == 1
        new_type = sub_types[0]
        # TODO: update foreign_keys/primary_keys in the type
        previous_fields = parent_schema.type().children()

        try:
            parent_stats = (await self.parent_marginals()).statistics()
            path = Path(self.dataset.transform().protobuf().spec.get_item.path)
            tab_multiplicity = parent_stats.nodes_statistics(path)[
                0
            ].multiplicity()
            properties = {MULTIPLICITY: str(tab_multiplicity)}
        except Exception:
            properties = {}
        if DATA in previous_fields.keys():
            previous_fields[DATA] = new_type
            new_type = sdt.Struct(fields=previous_fields)
        return schema(
            self.dataset,
            schema_type=new_type,
            privacy_unit_tracking_paths=None,
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
            properties=properties,
        )


class GetItem(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a get_item transform
    """

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        previous_ds = t.cast(Dataset, self.parent())
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        parent_schema = await self.parent_schema()

        async def get_item_func(batch: pa.RecordBatch) -> pa.Array:
            array = convert_record_batch(
                record_batch=batch, _type=parent_schema.type()
            )
            # VERY UGLY SHOULD BE REMOVED WHEN WE HAVE PROTECTED TYPE
            if DATA in parent_schema.type().children():
                old_arrays = array.flatten()
                idx_data = array.type.get_field_index(DATA)
                list_col_names = list(parent_schema.type().children().keys())
                list_idx = [
                    array.type.get_field_index(col) for col in list_col_names
                ]
                reordered_list_col_names = [
                    list_col_names[i]
                    for i in sorted(
                        range(len(list_idx)), key=lambda k: list_idx[k]
                    )
                ]
                array = old_arrays[idx_data]
                updated_array = get_items(
                    _type=parent_schema.data_type(),
                    array=array,
                    path=path,
                )
                old_arrays[idx_data] = updated_array
                return pa.StructArray.from_arrays(
                    old_arrays,
                    names=reordered_list_col_names,
                )

            updated_array = get_items(
                _type=parent_schema.data_type(),
                array=array,
                path=path,
            )
            if isinstance(updated_array, pa.StructArray):
                return updated_array
            return pa.StructArray.from_arrays(
                [updated_array],
                names=[path.to_strings_list()[0][-1]],
            )

        return await self.ensure_batch_correct(
            async_iterator=await previous_ds.async_to_arrow(
                batch_size=batch_size
            ),
            batch_size=batch_size,
            func_to_apply=get_item_func,
        )

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:
        sqlalchemy_metadata = await async_sa_metadata_from_dataset(
            t.cast(Dataset, self.parent())
        )
        queries: t.Dict[t.Tuple[str, ...], str] = {}
        table_path = Path(
            self.dataset.transform().protobuf().spec.get_item.path
        )
        curr_path = table_path.to_strings_list()[0]
        full_tablename = curr_path[1:] if curr_path[0] == DATA else curr_path
        sa_table_name = path_to_quoted_string(straight_path(full_tablename))
        sa_table = sqlalchemy_metadata.tables[sa_table_name]
        table_query = sa.select(sa_table)

        queries[()] = sqlalchemy_query_to_string(table_query)
        return queries

    async def size(self) -> st.Size:
        sizes = await self.parent_size()
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        new_stats = sizes.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return size_builder(dataset=self.dataset, statistics=new_stats[0])

    async def multiplicity(self) -> st.Multiplicity:
        multiplicities = await self.parent_multiplicity()
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        new_stats = multiplicities.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return multiplicity_builder(
            dataset=self.dataset, statistics=new_stats[0]
        )

    async def bounds(self) -> st.Bounds:
        bounds = await self.parent_bounds()
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        new_stats = bounds.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return bounds_builder(dataset=self.dataset, statistics=new_stats[0])

    async def marginals(self) -> st.Marginals:
        marginals = await self.parent_marginals()
        path = Path(self.dataset.transform().protobuf().spec.get_item.path)
        new_stats = marginals.statistics().nodes_statistics(path)
        assert len(new_stats) == 1
        return marg_builder(dataset=self.dataset, statistics=new_stats[0])


def get_items(array: pa.Array, path: st.Path, _type: st.Type) -> pa.Array:
    """Visitor selecting columns based on the type.
    The idea is that at each level,
    the filter for the array is computed, and for the union,
    we remove the fields that we want to filter among
    the columns
    """

    class ItemSelector(st.TypeVisitor):
        batch_array: pa.Array = array

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if len(path.sub_paths()) > 0:
                sub_path = path.sub_paths()[0]
                idx = array.type.get_field_index(sub_path.label())
                self.batch_array = get_items(
                    array=array.flatten()[idx],
                    path=sub_path,
                    _type=fields[sub_path.label()],
                )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            idx = self.batch_array.type.get_field_index(path.label())
            array = self.batch_array.flatten()(idx)
            if len(path.sub_paths()) == 0:
                self.batch_array = array
            else:
                self.batch_array = get_items(
                    array=array, path=path.sub_paths()[0], _type=type
                )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            if len(path.sub_paths()) == 0:
                self.batch_array = array
            else:
                sub_path = path.sub_paths()[0]
                idx = array.type.get_field_index(sub_path.label())
                self.batch_array = get_items(
                    array=array.flatten()[idx],
                    path=sub_path,
                    _type=fields[sub_path.label()],
                )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            pass

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            pass

    visitor = ItemSelector()
    _type.accept(visitor)
    return visitor.batch_array



from typing import Any, Callable, Literal, Optional, Tuple, Union

from numpy.typing import ArrayLike
import numpy as np

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import cluster
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any


class sk_birch(ExternalOpImplementation):
    _transform_id = "sklearn.SK_BIRCH"
    _signature = SarusSignature(
        SarusParameter(
            name="threshold",
            annotation=float,
            default=0.5,
        ),
        SarusParameter(
            name="branching_factor",
            annotation=int,
            default=50,
        ),
        SarusParameter(
            name="n_clusters",
            annotation=Optional[Union[int, BaseEstimator]],
            default=3,
        ),
        SarusParameter(
            name="compute_labels",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.Birch(**kwargs)


class sk_dbscan(ExternalOpImplementation):
    _transform_id = "sklearn.SK_DBSCAN"
    _signature = SarusSignature(
        SarusParameter(
            name="eps",
            annotation=float,
            default=0.5,
        ),
        SarusParameter(
            name="min_samples",
            annotation=int,
            default=5,
        ),
        SarusParameter(
            name="metric",
            annotation=Union[str, Callable],
            default="euclidean",
        ),
        SarusParameter(
            name="metric_params",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="algorithm",
            annotation=Literal["auto", "ball_tree", "kd_tree", "brute"],
            default="auto",
        ),
        SarusParameter(
            name="leaf_size",
            annotation=int,
            default=30,
        ),
        SarusParameter(
            name="p",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.DBSCAN(**kwargs)


class sk_feature_agglomeration(ExternalOpImplementation):
    _transform_id = "sklearn.SK_FEATURE_AGGLOMERATION"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=Optional[int],
            default=2,
        ),
        SarusParameter(
            name="affinity",
            annotation=Union[str, Callable],
            default="euclidean",
        ),
        SarusParameter(
            name="metric",
            annotation=Optional[Union[str, Callable]],
            default=None,
        ),
        SarusParameter(
            name="memory",
            annotation=Optional[Union[str, Any]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="connectivity",
            annotation=Optional[Union[ArrayLike, Callable]],
            default=None,
        ),
        SarusParameter(
            name="compute_full_tree",
            annotation=Union[Literal["auto"], bool],
            default="auto",
        ),
        SarusParameter(
            name="linkage",
            annotation=Literal["ward", "complete", "average", "single"],
            default="ward",
        ),
        SarusParameter(
            name="pooling_func",
            annotation=Callable,
            default=np.mean,
        ),
        SarusParameter(
            name="distance_threshold",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="compute_distances",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.FeatureAgglomeration(**kwargs)


class sk_kmeans(ExternalOpImplementation):
    _transform_id = "sklearn.SK_KMEANS"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=int,
            default=8,
        ),
        SarusParameter(
            name="init",
            annotation=Union[
                Literal["k-means++", "random"], Callable, ArrayLike
            ],
            default="k-means++",
        ),
        SarusParameter(
            name="n_init",
            annotation=Union[Literal["auto"], int],
            default=10,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=300,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=1e-4,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="copy_x",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="algorithm",
            annotation=Literal["lloyd", "elkan", "auto", "full"],
            default="lloyd",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.KMeans(**kwargs)


class sk_minibatch_kmeans(ExternalOpImplementation):
    _transform_id = "sklearn.SK_MINIBATCH_KMEANS"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=int,
            default=8,
        ),
        SarusParameter(
            name="init",
            annotation=Union[
                Literal["k-means++", "random"], Callable, ArrayLike
            ],
            default="k-means++",
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="batch_size",
            annotation=int,
            default=1024,
        ),
        SarusParameter(
            name="verbose",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="compute_labels",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="tol",
            annotation=float,
            default=0.0,
        ),
        SarusParameter(
            name="max_no_improvement",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="init_size",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="n_init",
            annotation=Union[Literal["auto"], int],
            default=3,
        ),
        SarusParameter(
            name="reassignment_ratio",
            annotation=float,
            default=0.01,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.MiniBatchKMeans(**kwargs)


class sk_mean_shift(ExternalOpImplementation):
    _transform_id = "sklearn.SK_MEAN_SHIFT"
    _signature = SarusSignature(
        SarusParameter(
            name="bandwidth",
            annotation=float,
            default=None,
        ),
        SarusParameter(
            name="seeds",
            annotation=Optional[ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="bin_seeding",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="min_bin_freq",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="cluster_all",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=300,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.MeanShift(**kwargs)


class sk_optics(ExternalOpImplementation):
    _transform_id = "sklearn.SK_OPTICS"
    _signature = SarusSignature(
        SarusParameter(
            name="min_samples",
            annotation=Union[int, float],
            default=5,
        ),
        SarusParameter(
            name="max_eps",
            annotation=float,
            default=np.inf,
        ),
        SarusParameter(
            name="metric",
            annotation=Union[
                Literal[
                    "cityblock",
                    "cosine",
                    "euclidean",
                    "l1",
                    "l2",
                    "manhattan",
                    "braycurtis",
                    "canberra",
                    "chebyshev",
                    "correlation",
                    "dice",
                    "hamming",
                    "jaccard",
                    "kulsinski",
                    "mahalanobis",
                    "minkowski",
                    "rogerstanimoto",
                    "russellrao",
                    "seuclidean",
                    "sokalmichener",
                    "sokalsneath",
                    "sqeuclidean",
                    "yule",
                ],
                Callable,
            ],
            default="minkowski",
        ),
        SarusParameter(
            name="p",
            annotation=float,
            default=2,
        ),
        SarusParameter(
            name="metric_params",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="cluster_method",
            annotation=Literal["xi", "dbscan"],
            default="xi",
        ),
        SarusParameter(
            name="eps",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="xi",
            annotation=float,
            default=0.5,
        ),
        SarusParameter(
            name="predecessor_correction",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="min_cluster_size",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="algorithm",
            annotation=Literal["auto", "ball_tree", "kd_tree", "brute"],
            default="auto",
        ),
        SarusParameter(
            name="leaf_size",
            annotation=int,
            default=30,
        ),
        SarusParameter(
            name="memory",
            annotation=Optional[Union[str, Any]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.OPTICS(**kwargs)


class sk_spectral_clustering(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SPECTRAL_CLUSTERING"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=int,
            default=8,
        ),
        SarusParameter(
            name="eigen_solver",
            annotation=Optional[Literal["arpack", "lobpcg", "amg"]],
            default=None,
        ),
        SarusParameter(
            name="n_components",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
        SarusParameter(
            name="n_init",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="gamma",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="affinity",
            annotation=Union[
                Literal[
                    "rbf",
                    "nearest_neighbors",
                    "precomputed",
                    "precomputed_nearest_neighbors",
                ],
                Callable,
            ],
            default="rbf",
        ),
        SarusParameter(
            name="n_neighbors",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="eigen_tol",
            annotation=Union[float, Literal["auto"]],
            default="auto",
        ),
        SarusParameter(
            name="assign_labels",
            annotation=Literal["kmeans", "discretize", "cluster_qr"],
            default="kmeans",
        ),
        SarusParameter(
            name="degree",
            annotation=int,
            default=3,
        ),
        SarusParameter(
            name="coef0",
            annotation=float,
            default=1.0,
        ),
        SarusParameter(
            name="kernel_params",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.SpectralClustering(**kwargs)


class sk_spectral_biclustering(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SPECTRAL_BICLUSTERING"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=Union[int, Tuple[int, int]],
            default=3,
        ),
        SarusParameter(
            name="method",
            annotation=Literal["bistochastic", "scale", "log"],
            default="bistochastic",
        ),
        SarusParameter(
            name="n_components",
            annotation=int,
            default=6,
        ),
        SarusParameter(
            name="n_best",
            annotation=int,
            default=3,
        ),
        SarusParameter(
            name="svd_method",
            annotation=Literal["randomized", "arpack"],
            default="randomized",
        ),
        SarusParameter(
            name="n_svd_vecs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="mini_batch",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="init",
            annotation=Union[Literal["k-means++", "random"], np.ndarray],
            default="k-means++",
        ),
        SarusParameter(
            name="n_init",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.SpectralBiclustering(**kwargs)


class sk_spectral_coclustering(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SPECTRAL_COCLUSTERING"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=int,
            default=3,
        ),
        SarusParameter(
            name="svd_method",
            annotation=Literal["randomized", "arpack"],
            default="randomized",
        ),
        SarusParameter(
            name="n_svd_vecs",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="mini_batch",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="init",
            annotation=Union[Literal["k-means++", "random"], np.ndarray],
            default="k-means++",
        ),
        SarusParameter(
            name="n_init",
            annotation=int,
            default=10,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.SpectralCoclustering(**kwargs)


class sk_affinity_propagation(ExternalOpImplementation):
    _transform_id = "sklearn.SK_AFFINITY_PROPAGATION"
    _signature = SarusSignature(
        SarusParameter(
            name="damping",
            annotation=float,
            default=0.5,
        ),
        SarusParameter(
            name="max_iter",
            annotation=int,
            default=200,
        ),
        SarusParameter(
            name="convergence_iter",
            annotation=int,
            default=15,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="preference",
            annotation=Optional[Union[ArrayLike, float]],
            default=None,
        ),
        SarusParameter(
            name="affinity",
            annotation=Literal["euclidean", "precomputed"],
            default="euclidean",
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="random_state",
            annotation=Optional[Union[int, np.random.RandomState]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.AffinityPropagation(**kwargs)


class sk_agglomerative_clustering(ExternalOpImplementation):
    _transform_id = "sklearn.SK_AGGLOMERATIVE_CLUSTERING"
    _signature = SarusSignature(
        SarusParameter(
            name="n_clusters",
            annotation=Optional[int],
            default=2,
        ),
        SarusParameter(
            name="affinity",
            annotation=Union[Literal["euclidean", "precomputed"], Callable],
            default="euclidean",
        ),
        SarusParameter(
            name="metric",
            annotation=Optional[
                Union[
                    Literal[
                        "euclidean",
                        "l1",
                        "l2",
                        "manhattan",
                        "cosine",
                        "precomputed",
                    ],
                    Callable,
                ]
            ],
            default=None,
        ),
        SarusParameter(
            name="memory",
            annotation=Optional[Union[str, Any]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="connectivity",
            annotation=Optional[Union[ArrayLike, Callable]],
            default=None,
        ),
        SarusParameter(
            name="compute_full_tree",
            annotation=Union[Literal["auto"], bool],
            default="auto",
        ),
        SarusParameter(
            name="linkage",
            annotation=Literal["ward", "complete", "average", "single"],
            default="ward",
        ),
        SarusParameter(
            name="distance_threshold",
            annotation=Optional[float],
            default=None,
        ),
        SarusParameter(
            name="compute_distances",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return cluster.AgglomerativeClustering(**kwargs)



import pyarrow as pa
import typing as t
import sarus_data_spec.typing as st
from sarus_data_spec.manager.ops.processor.standard.visitor_selector import (
    select_columns,
)
from sarus_data_spec.constants import DATA
from sarus_data_spec.arrow.pandas_utils import pandas_index_columns


def convert_record_batch(
    record_batch: pa.RecordBatch, _type: st.Type
) -> pa.Array:
    if str(_type.protobuf().WhichOneof("type")) not in ["struct", "union"]:
        return record_batch.column(0)
    record_schema = record_batch.schema
    fields = [record_schema.field(i) for i in range(len(record_schema.types))]
    return pa.StructArray.from_arrays(record_batch.columns, fields=fields)


async def fit_array_to_schema_type_async_gen(
    batch_async_iterator: t.AsyncIterator[pa.RecordBatch], schema_type: st.Type
) -> t.AsyncIterator[pa.RecordBatch]:
    """It makes sure that each pyarrow recordbatch is conform to the schema
    type provided.
    """
    async for batch in batch_async_iterator:
        if schema_type.has_admin_columns():
            assert DATA in batch.schema.names

        # retrieve pandas index that may not be present in the schema type
        field_names = list(batch.schema.names)

        # select columns present in the schema type.
        # it fails if columns in the schema are not in the array
        updated_array = select_columns(
            schema_type,
            convert_record_batch(record_batch=batch, _type=schema_type),
        )
        updated_batch = pa.RecordBatch.from_struct_array(updated_array)

        index_cols = [
            col
            for col in pandas_index_columns(batch.schema)
            if col not in list(updated_batch.schema.names)
        ]
        index_arrays = [
            batch.columns[field_names.index(col)] for col in index_cols
        ]
        if len(index_cols) > 0:
            # add index cols
            updated_field_names = list(updated_batch.schema.names)
            arrays = index_arrays + updated_batch.columns
            names = index_cols + updated_field_names
            new_struct_array = pa.StructArray.from_arrays(arrays, names)
            updated_batch = pa.RecordBatch.from_struct_array(new_struct_array)

        updated_batch = updated_batch.replace_schema_metadata(
            batch.schema.metadata
        )
        yield updated_batch



from typing import Any, List, Literal, Optional, Tuple, Union

from numpy.typing import ArrayLike

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn import compose
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any
    spmatrix = Any


class sk_column_transformer(ExternalOpImplementation):
    _transform_id = "sklearn.SK_COLUMN_TRANSFORMER"
    _signature = SarusSignature(
        SarusParameter(
            name="transformers",
            annotation=List[
                Tuple[
                    str,
                    Union[Literal["drop", "passthrough"], BaseEstimator],
                    Union[str, int, ArrayLike],
                ]
            ],
        ),
        SarusParameter(
            name="remainder",
            annotation=Union[Literal["drop", "passthrough"], BaseEstimator],
            default="drop",
        ),
        SarusParameter(
            name="sparse_threshold",
            annotation=float,
            default=0.3,
        ),
        SarusParameter(
            name="n_jobs",
            annotation=Optional[int],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="transformer_weights",
            annotation=Optional[dict],
            default=None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="verbose_feature_names_out",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return compose.ColumnTransformer(**kwargs)



from __future__ import annotations

from enum import Enum, auto
import itertools as it
import typing as t


class Operator(Enum):
    AND = auto()
    OR = auto()


class AcceptanceCondition:
    """Maintain a Sum Of Product (SOP) form of the condition.

    The Python representation is a list of sets.
    """

    products: t.List[t.Set[t.Union[AcceptanceCondition, ParameterKind]]]

    def __init__(
        self,
        left: t.Union[AcceptanceCondition, ParameterKind],
        right: t.Union[AcceptanceCondition, ParameterKind],
        operator: Operator,
    ):
        if isinstance(left, AcceptanceCondition):
            left_condition = t.cast(AcceptanceCondition, left)
            left_products = left_condition.products
        else:
            left_kind = t.cast(ParameterKind, left)
            left_products = [{left_kind}]

        if isinstance(right, AcceptanceCondition):
            right_condition = t.cast(AcceptanceCondition, right)
            right_products = right_condition.products
        else:
            right_kind = t.cast(ParameterKind, right)
            right_products = [{right_kind}]

        if operator == Operator.OR:
            self.products = left_products + right_products
        else:
            self.products = [
                set(right) | set(left)
                for right, left in it.product(right_products, left_products)
            ]

    def __repr__(self) -> str:
        return " ∨ ".join(
            [
                f"({' ∧ '.join([str(p) for p in product])})"
                for product in self.products
            ]
        )

    def __and__(
        self, other: t.Union[AcceptanceCondition, ParameterKind]
    ) -> AcceptanceCondition:
        return AcceptanceCondition(self, other, Operator.AND)

    def __or__(
        self, other: t.Union[AcceptanceCondition, ParameterKind]
    ) -> AcceptanceCondition:
        return AcceptanceCondition(self, other, Operator.OR)

    def isin(self, other: t.Union[AcceptanceCondition, ParameterKind]) -> bool:
        return is_accepted(self, other)


def is_accepted(
    accepted_kind: t.Union[AcceptanceCondition, ParameterKind],
    incoming_kind: t.Union[AcceptanceCondition, ParameterKind],
) -> bool:
    """Checks if the incoming parameter kind satisfies the
    acceptance condition.
    """
    if not isinstance(accepted_kind, ParameterKind):
        accepted_products = accepted_kind.products
    else:
        accepted_products = [{accepted_kind}]

    if not isinstance(incoming_kind, ParameterKind):
        incoming_products = incoming_kind.products
    else:
        incoming_products = [{incoming_kind}]

    return any(
        [
            accepted_prod.issubset(incoming_prod)
            for incoming_prod, accepted_prod in it.product(
                incoming_products, accepted_products
            )
        ]
    )


class ParameterKind(Enum):
    DATASET = auto()
    SCALAR = auto()
    PUP = auto()
    PUBLIC = auto()
    STATIC = auto()
    TRANSFORM = auto()

    def __and__(
        self, other: t.Union[AcceptanceCondition, ParameterKind]
    ) -> AcceptanceCondition:
        return AcceptanceCondition(self, other, Operator.AND)

    def __or__(
        self, other: t.Union[AcceptanceCondition, ParameterKind]
    ) -> AcceptanceCondition:
        return AcceptanceCondition(self, other, Operator.OR)

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    def isin(self, other: t.Union[AcceptanceCondition, ParameterKind]) -> bool:
        return is_accepted(self, other)


# Aliases
ParameterCondition = t.Union[AcceptanceCondition, ParameterKind]

DATASET = ParameterKind.DATASET
SCALAR = ParameterKind.SCALAR
STATIC = ParameterKind.STATIC
PUP = ParameterKind.PUP
PUP_DATASET = ParameterKind.DATASET & ParameterKind.PUP
DATASPEC = ParameterKind.SCALAR | ParameterKind.DATASET
TRANSFORM = ParameterKind.TRANSFORM



import logging
import time
import traceback
import typing as t

from sarus_data_spec import typing as st
from sarus_data_spec.constants import SCHEMA_TASK
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.cache_utils import lru_caching
from sarus_data_spec.manager.computations.local.base import LocalComputation
from sarus_data_spec.schema import Schema
from sarus_data_spec.status import DataSpecErrorStatus, error, ready

logger = logging.getLogger(__name__)


class SchemaComputation(LocalComputation[st.Schema]):
    """Class responsible to compute schemas"""

    task_name = SCHEMA_TASK

    async def prepare(self, dataspec: st.DataSpec) -> None:
        try:
            logger.info(f"STARTED SCHEMA {dataspec.uuid()}")
            start = time.perf_counter()
            schema = await self.computing_manager().async_schema_op(
                dataset=t.cast(Dataset, dataspec)
            )

        except DataSpecErrorStatus as exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(exception.relaunch),
                },
            )
            raise
        except Exception:
            error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    "relaunch": str(False),
                },
            )
            raise DataSpecErrorStatus((False, traceback.format_exc()))
        else:
            end = time.perf_counter()
            logger.info(
                f"FINISHED SCHEMA {dataspec.uuid()} ({end-start:.2f}s)"
            )
            ready(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={"uuid": schema.uuid()},
            )

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> st.Schema:
        return t.cast(
            Schema,
            dataspec.storage().referrable(properties["uuid"]),
        )

    @lru_caching("computation", use_first_arg=True)
    async def task_result(
        self, dataspec: st.DataSpec, **kwargs: t.Any
    ) -> st.Schema:
        return await super().task_result(dataspec, **kwargs)



# Copied and simplified from
# https://github.com/briancurtin/deprecation/blob/master/deprecation.py
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import functools
import textwrap
import typing as t
import warnings

#: Location where the details are added to a deprecated docstring
#: When set to ``"bottom"``, the details are appended to the end.
#: When set to ``"top"``, the details are inserted between the
#: summary line and docstring contents.
message_location = "bottom"


class DeprecatedWarning(DeprecationWarning):
    """A warning class for deprecated methods
    This is a specialization of the built-in :class:`DeprecationWarning`,
    adding parameters that allow us to get information into the __str__
    that ends up being sent through the :mod:`warnings` system.
    The attributes aren't able to be retrieved after the warning gets
    raised and passed through the system as only the class--not the
    instance--and message are what gets preserved.
    """

    def __init__(
        self,
        function: str,
        deprecated_in: str,
        removed_in: str,
        details: str = "",
    ):
        # NOTE: The docstring only works for this class if it appears up
        # near the class name, not here inside __init__
        self.function = function
        self.deprecated_in = deprecated_in
        self.removed_in = removed_in
        self.details = details
        super(DeprecatedWarning, self).__init__(
            function, deprecated_in, removed_in, details
        )

    def __str__(self) -> str:
        # Use a defaultdict to give us the empty string
        # when a part isn't included.
        parts: t.DefaultDict[str, t.Any] = collections.defaultdict(str)
        parts["function"] = self.function

        if self.deprecated_in:
            parts["deprecated"] = " as of %s" % self.deprecated_in
        if self.removed_in:
            parts["removed"] = " and will be removed in {}".format(
                self.removed_in
            )
        if any([self.deprecated_in, self.removed_in, self.details]):
            parts["period"] = "."
        if self.details:
            parts["details"] = " %s" % self.details

        return (
            "%(function)s is deprecated%(deprecated)s%(removed)s"
            "%(period)s%(details)s" % (parts)
        )


def deprecated(
    deprecated_in: str,
    removed_in: str,
    details: str = "",
) -> t.Callable:
    """Decorate a function to signify its deprecation
    This function wraps a method that will soon be removed and does two things:
        * The docstring of the method will be modified to include a notice
          about deprecation, e.g., "Deprecated since 0.9.11. Use foo instead."
        * Raises a :class:`~deprecation.DeprecatedWarning`
          via the :mod:`warnings` module, which is a subclass of the built-in
          :class:`DeprecationWarning`. Note that built-in
          :class:`DeprecationWarning`s are ignored by default, so for users
          to be informed of said warnings they will need to enable them--see
          the :mod:`warnings` module documentation for more details.
    :param deprecated_in: The version at which the decorated method is
                          considered deprecated. This will usually be the
                          next version to be released when the decorator is
                          added. The default is **None**, which effectively
                          means immediate deprecation. If this is not
                          specified, then the `removed_in` and
                          `current_version` arguments are ignored.
    :param removed_in: The default is **None**,
                       specifying that the function is not currently planned
                       to be removed.
                       Note: This parameter cannot be set to a value if
                       `deprecated_in=None`.
    :param details: Extra details to be added to the method docstring and
                    warning. For example, the details may point users to
                    a replacement method, such as "Use the foo_bar
                    method instead". By default there are no details.
    """
    # Only warn when it's appropriate. There may be cases when it makes sense
    # to add this decorator before a formal deprecation period begins.
    # In CPython, PendingDeprecatedWarning gets used in that period,
    # so perhaps mimick that at some point.

    def _function_wrapper(function: t.Callable) -> t.Callable:
        # Everything *should* have a docstring, but just in case...
        existing_docstring = function.__doc__ or ""

        # The various parts of this decorator being optional makes for
        # a number of ways the deprecation notice could go. The following
        # makes for a nicely constructed sentence with or without any
        # of the parts.
        parts = {
            "deprecated_in": " %s" % deprecated_in if deprecated_in else "",
            "removed_in": "\n   This will be removed in {}.".format(removed_in)
            if removed_in
            else "",
            "details": " %s" % details if details else "",
        }

        deprecation_note = (
            ".. deprecated::{deprecated_in}" "{removed_in}{details}".format(
                **parts
            )
        )

        # default location for insertion of deprecation note
        loc = 1

        # split docstring at first occurrence of newline
        string_list = existing_docstring.split("\n", 1)

        if len(string_list) > 1:
            # With a multi-line docstring, when we modify
            # existing_docstring to add our deprecation_note,
            # if we're not careful we'll interfere with the
            # indentation levels of the contents below the
            # first line, or as PUP 257 calls it, the summary
            # line. Since the summary line can start on the
            # same line as the """, dedenting the whole thing
            # won't help. Split the summary and contents up,
            # dedent the contents independently, then join
            # summary, dedent'ed contents, and our
            # deprecation_note.

            # in-place dedent docstring content
            string_list[1] = textwrap.dedent(string_list[1])

            # we need another newline
            string_list.insert(loc, "\n")

            # change the message_location if we add to end of docstring
            # do this always if not "top"
            if message_location != "top":
                loc = 3

        # insert deprecation note and dual newline
        string_list.insert(loc, deprecation_note)
        string_list.insert(loc, "\n\n")
        function.__doc__ = "".join(string_list)

        @functools.wraps(function)
        def _inner(*args: t.Any, **kwargs: t.Any) -> t.Any:
            the_warning = DeprecatedWarning(
                function.__name__, deprecated_in, removed_in, details
            )
            warnings.warn(
                the_warning, category=DeprecationWarning, stacklevel=2
            )
            return function(*args, **kwargs)

        return _inner

    return _function_wrapper


def fail_if_not_removed(method: t.Callable) -> t.Callable:
    """Decorate a test method to track removal of deprecated code
    This decorator catches :class:`~deprecation.UnsupportedWarning`
    warnings that occur during testing and causes unittests to fail,
    making it easier to keep track of when code should be removed.
    :raises: :class:`AssertionError` if an
             :class:`~deprecation.UnsupportedWarning`
             is raised while running the test method.
    """

    # NOTE: Unless this is named test_inner, nose won't work
    # properly.See
    @functools.wraps(method)
    def test_inner(*args: t.Any, **kwargs: t.Any) -> t.Any:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            rv = method(*args, **kwargs)
        for warning in caught_warnings:
            if warning.category == DeprecatedWarning:
                raise AssertionError(
                    (
                        "%s uses a function that should be removed: %s"
                        % (method, str(warning.message))
                    )
                )
        return rv

    return test_inner



from __future__ import annotations

from typing import List, Optional, Type, cast
import json

from sarus_data_spec.base import Referring
from sarus_data_spec.constants import (
    IS_PUBLIC,
    IS_SYNTHETIC,
    PRIVACY_LIMIT,
    PUP_TOKEN,
    SALT,
)
from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class VariantConstraint(Referring[sp.VariantConstraint]):
    def __init__(
        self, protobuf: sp.VariantConstraint, store: bool = True
    ) -> None:
        self._referred = {protobuf.dataspec}
        super().__init__(protobuf=protobuf, store=store)

    def prototype(self) -> Type[sp.VariantConstraint]:
        """Return the type of the underlying protobuf."""
        return sp.VariantConstraint

    def constraint_kind(self) -> st.ConstraintKind:
        return st.ConstraintKind(self.protobuf().constraint_kind)

    def required_context(self) -> List[str]:
        return list(self.protobuf().required_context)

    def privacy_limit(self) -> Optional[st.PrivacyLimit]:
        """Instantiate the privacy limit from the protobuf properties."""
        dump = self.protobuf().properties.get(PRIVACY_LIMIT, None)
        if dump is None:
            return None

        # in JSON keys are converted to str, we reconvert to float here
        data_epsilon_dict = {float(k): v for k, v in json.loads(dump).items()}
        return DeltaEpsilonLimit(data_epsilon_dict)

    def dataspec(self) -> st.DataSpec:
        return cast(
            st.DataSpec, self.storage().referrable(self.protobuf().dataspec)
        )


# Builders


def public_constraint(
    dataspec: st.DataSpec, is_public: bool = True
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is Public."""
    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.PUBLIC,
            properties={
                IS_PUBLIC: str(is_public),
            },
        )
    )


def dp_constraint(
    dataspec: st.DataSpec,
    required_context: List[str],
    privacy_limit: st.PrivacyLimit,
    salt: Optional[int] = None,
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is DP."""
    properties = {
        PRIVACY_LIMIT: json.dumps(privacy_limit.delta_epsilon_dict()),
    }
    if salt is not None:
        properties[SALT] = str(salt)

    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.DP,
            required_context=required_context,
            properties=properties,
        )
    )


def pup_constraint(
    dataspec: st.DataSpec,
    token: str,
    required_context: List[str],
    privacy_limit: st.PrivacyLimit,
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is PUP."""
    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.PUP,
            required_context=required_context,
            properties={
                PRIVACY_LIMIT: json.dumps(privacy_limit.delta_epsilon_dict()),
                PUP_TOKEN: token,
            },
        )
    )


def pup_for_rewriting_constraint(
    dataspec: st.DataSpec,
    token: str,
    required_context: List[str],
    privacy_limit: st.PrivacyLimit,
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is PUP."""
    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.PUP_FOR_REWRITING,
            required_context=required_context,
            properties={
                PRIVACY_LIMIT: json.dumps(privacy_limit.delta_epsilon_dict()),
                PUP_TOKEN: token,
            },
        )
    )


def syn_constraint(
    dataspec: st.DataSpec,
    required_context: Optional[List[str]] = None,
    is_synthetic: bool = True,
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is synthetic."""
    if required_context is None:
        required_context = []
    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.SYNTHETIC,
            required_context=required_context,
            properties={
                IS_SYNTHETIC: str(is_synthetic),
            },
        )
    )


def mock_constraint(
    dataspec: st.DataSpec,
    required_context: Optional[List[str]] = None,
) -> VariantConstraint:
    """Create a variant constraint specifying a dataspec is a mock."""
    if required_context is None:
        required_context = []
    return VariantConstraint(
        sp.VariantConstraint(
            dataspec=dataspec.uuid(),
            constraint_kind=sp.ConstraintKind.MOCK,
            required_context=required_context,
        )
    )



from typing import Dict, List, cast
import warnings

try:
    from scipy.optimize import bisect
except ModuleNotFoundError:
    warnings.warn("Scipy not available")

from sarus_data_spec.dataset import Dataset
import sarus_data_spec.typing as st


async def differentiated_sampling_sizes(
    dataset: st.Dataset,
) -> Dict[st.Path, int]:
    """Get the sampling rates for each table"""
    parent_ds = cast(Dataset, dataset.parents()[0][0])
    parent_size = await parent_ds.manager().async_size(parent_ds)
    assert parent_size

    transform = dataset.transform().protobuf().spec.differentiated_sample
    if transform.HasField("fraction"):
        sampled_size = parent_size.statistics().size() * transform.fraction
    else:
        sampled_size = transform.size

    return await differentiated_sampling_sizes_bisection(
        parent_ds, int(sampled_size)
    )


async def differentiated_sampling_sizes_bisection(
    dataset: st.Dataset, sampled_size: int
) -> Dict[st.Path, int]:
    """Get the sampling rates for each table"""
    size = await dataset.manager().async_size(dataset)
    assert size
    size_stat = size.statistics()
    schema = await dataset.manager().async_schema(dataset)
    private_table_paths = schema.private_tables()
    public_table_paths = schema.public_tables()

    # test if min_table_size is not too high
    private_table_path_sizes = {
        table_path: size_stat.nodes_statistics(table_path)[0].size()
        for table_path in private_table_paths
    }
    public_table_path_sizes = {
        table_path: size_stat.nodes_statistics(table_path)[0].size()
        for table_path in public_table_paths
    }

    def new_size(max_table_size: float) -> int:
        return int(
            sum(
                [
                    min(
                        int(max_table_size),
                        table_path_size,
                    )
                    for table_path_size in private_table_path_sizes.values()
                ]
            )
            - sampled_size
        )

    private_size = sum(private_table_path_sizes.values())
    if private_size < sampled_size:
        # if already below sample size
        return {**private_table_path_sizes, **public_table_path_sizes}

    max_table_size = bisect(
        new_size, 0, private_size, xtol=len(private_table_path_sizes)
    )
    return {
        **{
            table_path: min(
                int(max_table_size),
                table_path_size,
            )
            for table_path, table_path_size in private_table_path_sizes.items()
        },
        **public_table_path_sizes,
    }


def sync_bisection(
    size_stat: st.Statistics, table_paths: List[st.Path], sampled_size: int
) -> Dict[st.Path, int]:
    table_path_sizes = {
        table_path: size_stat.nodes_statistics(table_path)[0].size()
        for table_path in table_paths
    }

    def new_size(max_table_size: float) -> int:
        return int(
            sum(
                [
                    min(
                        int(max_table_size),
                        table_path_size,
                    )
                    for table_path_size in table_path_sizes.values()
                ]
            )
            - sampled_size
        )

    if size_stat.size() < sampled_size:
        # if already below sample size
        return table_path_sizes

    max_table_size = bisect(
        new_size, 0, size_stat.size(), xtol=len(table_paths)
    )
    return {
        table_path: min(
            int(max_table_size),
            table_path_size,
        )
        for table_path, table_path_size in table_path_sizes.items()
    }



from typing import Any, List, Optional

import numpy as np

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC, STATIC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    import shap
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class shap_hclust(ExternalOpImplementation):
    _transform_id = "shap.SHAP_HCLUST"
    _signature = SarusSignature(
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="y",
            annotation=Optional[Any],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="linkage",
            annotation=str,
            default="single",
        ),
        SarusParameter(
            name="metric",
            annotation=str,
            default="auto",
        ),
        SarusParameter(
            name="random_state",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.hclust(**kwargs)


class shap_hclust_ordering(ExternalOpImplementation):
    _transform_id = "shap.SHAP_HCLUST_ORDERING"
    _signature = SarusSignature(
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="metric",
            annotation=str,
            default="sqeuclidean",
        ),
        SarusParameter(
            name="anchor_first",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.hclust_ordering(**kwargs)


class shap_partition_tree(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PARTITION_TREE"
    _signature = SarusSignature(
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="metric",
            annotation=str,
            default="correlation",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.partition_tree(**kwargs)


class shap_partition_tree_shuffle(ExternalOpImplementation):
    _transform_id = "shap.SHAP_PARTITION_TREE_SHUFFLE"
    _signature = SarusSignature(
        SarusParameter(
            name="indexes",
            annotation=np.array,
        ),
        SarusParameter(
            name="index_mask",
            annotation=np.array,
        ),
        SarusParameter(
            name="partition_tree",
            annotation=np.array,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.partition_tree_shuffle(**kwargs)


class shap_delta_minimization_order(ExternalOpImplementation):
    _transform_id = "shap.SHAP_DELTA_MINIMIZATION_ORDER"
    _signature = SarusSignature(
        SarusParameter(
            name="all_masks",
            annotation=Any,
        ),
        SarusParameter(
            name="max_swap_size",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="num_passes",
            annotation=int,
            default=2,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.delta_minimization_order(**kwargs)


class shap_approximate_interactions(ExternalOpImplementation):
    _transform_id = "shap.SHAP_APPROXIMATE_INTERACTIONS"
    _signature = SarusSignature(
        SarusParameter(
            name="index",
            annotation=Any,
        ),
        SarusParameter(
            name="shap_values",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="feature_names",
            annotation=Optional[List[str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.approximate_interactions(**kwargs)


class shap_potential_interactions(ExternalOpImplementation):
    _transform_id = "shap.SHAP_POTENTIAL_INTERACTIONS"
    _signature = SarusSignature(
        SarusParameter(
            name="shap_values_column",
            annotation=Any,
            condition=STATIC,
        ),
        SarusParameter(
            name="shap_values_matrix",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        args = signature.collect_args()
        return shap.utils.potential_interactions(*args)


class shap_sample(ExternalOpImplementation):
    _transform_id = "shap.SHAP_SAMPLE"
    _signature = SarusSignature(
        SarusParameter(
            name="X",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="nsamples",
            annotation=int,
            default=100,
        ),
        SarusParameter(
            name="random_state",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return shap.utils.sample(**kwargs)


class shap_convert_name(ExternalOpImplementation):
    _transform_id = "shap.SHAP_CONVERT_NAME"
    _signature = SarusSignature(
        SarusParameter(
            name="ind",
            annotation=Any,
        ),
        SarusParameter(
            name="shap_values",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="input_names",
            annotation=Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        args = signature.collect_args()
        return shap.utils.convert_name(*args)
